"""Base support for CI-based integrations.

This can be used by both internal and external integrations to more easily
integrate with CI solutions for Review Board.

Version Added:
    3.1
"""

from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import List, TYPE_CHECKING, cast

from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured
from django.db import IntegrityError, transaction
from django.utils import timezone
from djblets.avatars.services import URLAvatarService
from djblets.secrets.token_generators import token_generator_registry
from djblets.siteconfig.models import SiteConfiguration
from reviewboard.admin.server import get_server_url
from reviewboard.avatars import avatar_services
from reviewboard.diffviewer.models import DiffSet
from reviewboard.extensions.hooks import SignalHook
from reviewboard.integrations.base import Integration
from reviewboard.integrations.models import IntegrationConfig
from reviewboard.reviews.models import StatusUpdate
from reviewboard.reviews.signals import (review_request_published,
                                         status_update_request_run)
from reviewboard.webapi.models import WebAPIToken

from rbintegrations.baseci.errors import CIBuildError

if TYPE_CHECKING:
    from collections.abc import Sequence

    from djblets.util.typing import JSONDict

    from reviewboard.changedescs.models import ChangeDescription
    from reviewboard.reviews.models import ReviewRequest
    from reviewboard.scmtools.models import Repository
    from reviewboard.site.models import LocalSite


logger = logging.getLogger(__name__)


_has_create_for_integration = hasattr(StatusUpdate.objects,
                                      'create_for_integration')


@dataclass
class BuildPrepData:
    """Data used in the preparation for a build.

    This contains state calculated by the base class and by a particular
    integration for use in configuring and running a build. It's also used
    to help create and manage :term:`status updates`.

    Version Added:
        3.1
    """

    #: The list of matching configurations for the build.
    #:
    #: Integrations can modify this list to narrow down the configurations
    #: further, if needed.
    #:
    #: By default, it will contain all configurations that match the
    #: integration configuration's conditions and :term:`Local Site`.
    #:
    #: Type:
    #:     list of reviewboard.integrations.models.IntegrationConfig
    configs: list[IntegrationConfig]

    #: The DiffSet being applied and tested against.
    #:
    #: Type:
    #:     reviewboard.diffviewer.models.DiffSet
    diffset: DiffSet

    #: The review request being tested against.
    #:
    #: Type:
    #:     reviewboard.reviews.models.review_request.ReviewRequest
    review_request: ReviewRequest

    #: The user owning the StatusUpdate and any reviews.
    #:
    #: Type:
    #:     django.contrib.auth.models.User
    user: User

    #: The optional Change Description a StatusUpdate should associate with.
    #:
    #: Type:
    #:     reviewboard.changedescs.models.ChangeDescription
    changedesc: (ChangeDescription | None) = None

    #: Extra state set by the integration.
    #:
    #: The integration can use this to store and access any custom state
    #: needed for the build steps.
    #:
    #: Type:
    #:     dict
    extra_state: JSONDict = field(default_factory=dict)

    @property
    def local_site(self) -> LocalSite:
        """The Local Site these builds will be performed on.

        Type:
            reviewboard.site.models.LocalSite
        """
        return self.review_request.local_site

    @property
    def repository(self) -> Repository:
        """The repository the diff applies to.

        Type:
            reviewboard.scmtools.models.Repository
        """
        return self.review_request.repository

    @property
    def server_url(self) -> str:
        """The URL to the root of the server or Local Site.

        Type:
            str
        """
        return get_server_url(local_site=self.local_site)

    @property
    def api_token(self) -> WebAPIToken:
        """The API token used to communicate with Review Board.

        If one doesn't already exist for the user, one will be generated
        and stored.

        Type:
            reviewboard.webapi.models.WebAPIToken
        """
        user = self.user
        local_site = self.local_site

        try:
            return user.webapi_tokens.filter(local_site=local_site)[0]
        except IndexError:
            token_generator = token_generator_registry.get_default()

            return WebAPIToken.objects.generate_token(
                user,
                local_site=local_site,
                auto_generated=True,
                token_generator_id=token_generator.token_generator_id,
                token_info={'token_type': 'rbp'},
            )


class BaseCIIntegration(Integration):
    """Base class for CI integrations.

    This manages the process of listening for new review requests or manual
    requests to run a build, setting up the necessary state (including any
    :term:`status updates`), and performing the build.

    Subclasses only need to implement:

    * :py:meth:`start_build`
    * :py:meth:`prepare_builds` (optional)

    And set some attributes:

    * :py:attr:`status_update_service_id`
    * :py:attr:`bot_username`
    * :py:attr:`bot_user_first_name`
    * :py:attr:`bot_user_last_name`
    """

    #: The ID of the service used for any status updates.
    #:
    #: This must be set by subclasses.
    #:
    #: Type:
    #:     str
    status_update_service_id: (str | None) = None

    #: The username for the bot user to create/use for any status updates.
    #:
    #: This must be set by subclasses.
    #:
    #: Type:
    #:     str
    bot_username: (str | None) = None

    #: The first name for the bot user.
    #:
    #: This must be set by subclasses.
    #:
    #: Type:
    #:     str
    bot_user_first_name: (str | None) = None

    #: The last name for the bot user.
    #:
    #: This must be set by subclasses.
    #:
    #: Type:
    #:     str
    bot_user_last_name: (str | None) = None

    def initialize(self) -> None:
        """Initialize the integration.

        This will begin listening for any signals needed to trigger builds or
        build preparation.
        """
        for attr_name in ('name',
                          'status_update_service_id',
                          'bot_username',
                          'bot_user_first_name',
                          'bot_user_last_name'):
            if not getattr(self, attr_name, None):
                raise ImproperlyConfigured(
                    f'{type(self).__name__} must set the "{attr_name}" '
                    f'attribute.'
                )

        SignalHook(self,
                   review_request_published,
                   self._on_review_request_published)
        SignalHook(self,
                   status_update_request_run,
                   self._on_status_update_request_run)

    def prepare_builds(
        self,
        prep_data: BuildPrepData,
    ) -> bool:
        """Prepare for builds.

        Subclasses can set this to set or modify any state in ``prep_data``
        before builds using this state are conducted.

        This should also return a boolean indicating whether to proceed with
        builds.

        Args:
            prep_data (BuildPrepData):
                The builds preparation data to modify or use.

        Returns:
            bool:
            ``True`` if the builds should proceed. ``False`` if builds should
            be skipped.
        """
        return True

    def start_build(
        self,
        *,
        prep_data: BuildPrepData,
        config: IntegrationConfig,
        status_update: StatusUpdate,
    ) -> None:
        """Start a new build.

        This will be called any time a build is being automatically started
        through a new published diff, or manually run via a
        :term:`status update`.

        Subclasses must override this to trigger a build. This should be a
        fast operation, and the implementation must not wait around for the
        results of a build. Instead, builds should communicate their status
        with Review Board via the API.

        ``status_update`` will already be in the correct running state.
        Subclasses should set this to a failure state if triggering a build
        fails. If an unexpected exception is raised, it will be automatically
        placed in an error state.

        A subclass may also set any relevant information (such as a URL) on the
        ``status_update``.

        Args:
            prep_data (BuildPrepData):
                The builds preparation data containing information for the
                build.

            config (reviewboard.integrations.models.IntegrationConfig):
                The configuration for the integration triggering this build.

            status_update (reviewboard.reviews.models.status_update.
                           StatusUpdate):
                The status update indicating the status of this build.

        Raises:
            rbintegrations.baseci.errors.CIBuildError:
                There was an error invoking a build.

                Subclasses can raise this to set information for the
                :py:class:`~reviewboard.reviews.models.status_update.
                StatusUpdate`.
        """
        raise NotImplementedError

    def get_or_create_api_token(
        self,
        user: User,
        local_site: LocalSite | None,
    ) -> WebAPIToken:
        """Return or create an API token used to communicate with Review Board.

        Args:
            user (django.contrib.auth.models.User):
                The user that will own the token.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site that the token will be restricted to.

        Returns:
            reviewboard.webapi.models.WebAPIToken:
            The resulting API token.
        """
        try:
            return user.webapi_tokens.filter(local_site=local_site)[0]
        except IndexError:
            token_generator = token_generator_registry.get_default()

            return WebAPIToken.objects.generate_token(
                user,
                local_site=local_site,
                auto_generated=True,
                token_generator_id=token_generator.token_generator_id,
                token_info={'token_type': 'rbp'})

    def get_or_create_user(self) -> User:
        """Return or create a bot user for this integration.

        The bot user will be associated with any status updates and reviews
        for the builds. It should also be used for any communication between
        the service or program running the build, and Review Board's API.
        """
        username = self.bot_username

        try:
            return User.objects.get(username=username)
        except User.DoesNotExist:
            logger.info('Creating new user "%s" for CI integration "%s"',
                        username, self.name)

            siteconfig = SiteConfiguration.objects.get_current()
            noreply_email = siteconfig.get('mail_default_from')

            with transaction.atomic():
                try:
                    user = User.objects.create(
                        username=username,
                        email=noreply_email,
                        first_name=self.bot_user_first_name,
                        last_name=self.bot_user_last_name)
                except IntegrityError:
                    # Another process/thread beat us to it.
                    return User.objects.get(username=username)

                profile = user.get_profile()
                profile.should_send_email = False
                profile.save(update_fields=('should_send_email',))

                # Set up an avatar for this user, using the integration's
                # icons.
                avatar_service_id = URLAvatarService.avatar_service_id

                if avatar_services.is_enabled(avatar_service_id):
                    avatar_service = avatar_services.get_avatar_service(
                        avatar_service_id)
                    avatar_service.setup(user, self.icon_static_urls)

                return user

    def get_matching_configs(
        self,
        review_request: ReviewRequest,
    ) -> list[IntegrationConfig]:
        """Return configurations that match the review request and integration.

        This will return all configurations found in the database that are
        a match for the review request and its :term:`Local Site`.

        Subclasses can override this to further filter configurations, or
        filter this in :py:meth:`prepare_builds`.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request being matched.

        Returns:
            list of reviewboard.integrations.models.IntegrationConfig:
            The list of matched configurations.
        """
        config_form_cls = self.config_form_cls
        local_site = review_request.local_site
        configs = cast(List[IntegrationConfig],
                       self.get_configs(local_site=local_site))

        return [
            config
            for config in configs
            if config.match_conditions(form_cls=config_form_cls,
                                       review_request=review_request)
        ]

    def update_status(
        self,
        status_update: StatusUpdate,
        *,
        state: (str | None) = None,
        description: (str | None) = None,
        url: (str | None) = None,
        url_text: (str | None) = None,
        save: bool = True,
        extra_fields: (Sequence[str] | None) = None,
    ) -> None:
        """Update fields on a StatusUpdate.

        This will update only the provided fields, along with the timestamp,
        and optionally saving them to the database.

        This should be used when subclasses need to modify a
        :py:class:`~reviewboard.reviews.models.status_update.StatusUpdate`.

        Version Changed:
            5.0:
            Added the ``extra_fields`` argument.

        Args:
            status_update (reviewboard.reviews.models.status_update.
                           StatusUpdate):
                The status update to modify.

            state (str, optional):
                A new state string.

            description (str, optional):
                A new description.

            url (str, optional):
                A new URL for build information.

            url_text (str, optional):
                New text for a URL for build information.

            save (bool, optional):
                Whether to save the resulting
                :py:class:`~reviewboard.reviews.models.status_update.
                StatusUpdate`.

            extra_fields (list of str, optional):
                Any extra fields to include when saving the result.

                Version Added:
                    5.0
        """
        fields: list[str] = []

        if state is not None and state != status_update.state:
            status_update.state = state
            fields.append('state')

        if (description is not None and
            description != status_update.description):
            status_update.description = description
            fields.append('description')

        if url is not None and url != status_update.url:
            status_update.url = url
            fields.append('url')

        if url_text is not None and url_text != status_update.url_text:
            status_update.url_text = url_text
            fields.append('url_text')

        if extra_fields is not None:
            fields.extend(extra_fields)

        if fields:
            status_update.timestamp = timezone.now()
            fields.append('timestamp')

            if save:
                if status_update.pk:
                    status_update.save(update_fields=fields)
                else:
                    status_update.save()

    def set_waiting(
        self,
        status_update: StatusUpdate,
        *,
        description: (str | None) = None,
        url: (str | None) = None,
        url_text: (str | None) = None,
        save: bool = True,
    ) -> None:
        """Set a status update to "waiting to run" mode.

        Args:
            status_update (reviewboard.reviews.models.status_update.
                           StatusUpdate):
                The status update to modify.

            description (str, optional):
                A new description. If not provided, this will default to
                "waiting to run."

            url (str, optional):
                A new URL for build information.

            url_text (str, optional):
                New text for a URL for build information.

            save (bool, optional):
                Whether to save the resulting
                :py:class:`~reviewboard.reviews.models.status_update.
                StatusUpdate`.
        """
        self.update_status(status_update,
                           state=status_update.NOT_YET_RUN,
                           description=description or 'waiting to run.',
                           url=url,
                           url_text=url_text,
                           save=save)

    def set_starting(
        self,
        status_update: StatusUpdate,
        *,
        description: (str | None) = None,
        url: (str | None) = None,
        url_text: (str | None) = None,
        save: bool = True,
    ) -> None:
        """Set a status update to "starting build" mode.

        Args:
            status_update (reviewboard.reviews.models.status_update.
                           StatusUpdate):
                The status update to modify.

            description (str, optional):
                A new description. If not provided, this will default to
                "starting build..."

            url (str, optional):
                A new URL for build information.

            url_text (str, optional):
                New text for a URL for build information.

            save (bool, optional):
                Whether to save the resulting
                :py:class:`~reviewboard.reviews.models.status_update.
                StatusUpdate`.
        """
        self.update_status(status_update,
                           state=status_update.PENDING,
                           description=description or 'starting build...',
                           url=url,
                           url_text=url_text,
                           save=save)

    def set_error(
        self,
        status_update: StatusUpdate,
        *,
        description: (str | None) = None,
        url: (str | None) = None,
        url_text: (str | None) = None,
        save: bool = True,
    ) -> None:
        """Set a status update to "internal error" mode.

        Args:
            status_update (reviewboard.reviews.models.status_update.
                           StatusUpdate):
                The status update to modify.

            description (str, optional):
                A new description. If not provided, this will default to
                "internal error."

            url (str, optional):
                A new URL for build information.

            url_text (str, optional):
                New text for a URL for build information.

            save (bool, optional):
                Whether to save the resulting
                :py:class:`~reviewboard.reviews.models.status_update.
                StatusUpdate`.
        """
        self.update_status(status_update,
                           state=status_update.ERROR,
                           description=description or 'internal error.',
                           url=url,
                           url_text=url_text,
                           save=save)

    def _on_review_request_published(
        self,
        sender: type[ReviewRequest],
        review_request: ReviewRequest,
        changedesc: (ChangeDescription | None) = None,
        **kwargs,
    ) -> None:
        """Handle when a review request is published.

        Args:
            sender (type):
                The sender of the signal.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request which was published.

            changedesc (reviewboard.changedescs.models.ChangeDescription,
                        optional):
                The change description associated with the publish.

            **kwargs (dict):
                Additional keyword arguments.
        """
        # Don't build any review request that don't have matching configs.
        matching_configs = self.get_matching_configs(review_request)

        if not matching_configs:
            return

        # Don't build any review requests that don't include diffs.
        diffset = review_request.get_latest_diffset()

        if not diffset:
            return

        # If this was an update to a review request, make sure that there was a
        # diff update in it.
        if changedesc is not None:
            fields_changed = changedesc.fields_changed

            if ('diff' not in fields_changed or
                'added' not in fields_changed['diff']):
                return

        # Build the preparation data and find out if the subclass wants to
        # perform this build.
        user = self.get_or_create_user()
        prep_data = BuildPrepData(
            changedesc=changedesc,
            configs=matching_configs,
            diffset=diffset,
            review_request=review_request,
            user=user)

        if not self.prepare_builds(prep_data):
            return

        # Create a StatusUpdate for each matching configuration, and start
        # a build.
        status_update_service_id = self.status_update_service_id
        status_update_name = self.name

        for config in matching_configs:
            run_manually = config.get('run_manually')
            timeout_secs = config.get('timeout')

            if _has_create_for_integration:
                # Review Board >= 5.0.3
                status_update = StatusUpdate.objects.create_for_integration(
                    self,
                    config=config,
                    service_id=status_update_service_id,
                    user=user,
                    summary=str(status_update_name),
                    review_request=review_request,
                    change_description=changedesc,
                    can_retry=True,
                    timeout=timeout_secs,
                    starting_description='starting build...')
            else:
                # Review Board <= 5.0.2
                status_update = StatusUpdate(
                    service_id=status_update_service_id,
                    user=user,
                    summary=status_update_name,
                    review_request=review_request,
                    change_description=changedesc,
                    extra_data={
                        '__integration_config_id': config.pk,
                        'can_retry': True,
                    })

                # These will both save the StatusUpdate.
                if run_manually:
                    self.set_waiting(status_update)
                else:
                    self.set_starting(status_update)

            if not run_manually:
                self._run_start_build(prep_data=prep_data,
                                      config=config,
                                      status_update=status_update)

    def _on_status_update_request_run(
        self,
        sender: type[StatusUpdate],
        status_update: StatusUpdate,
        config: (IntegrationConfig | None) = None,
        **kwargs,
    ) -> None:
        """Handle a request to run or rerun a build.

        Args:
            sender (type):
                The sender of the signal.

            status_update (reviewboard.reviews.models.StatusUpdate):
                The status update being run.

            **kwargs (dict):
                Any additional keyword arguments.
        """
        assert self.status_update_service_id

        service_id: str = status_update.service_id

        if not service_id.startswith(self.status_update_service_id):
            # Ignore anything that's not this CI integration.
            return

        review_request: ReviewRequest = status_update.review_request
        changedesc = status_update.change_description

        if config is None:
            # This is a StatusUpdate created with Review Board < 5.0.3 or
            # rbintegrations < 3.1.
            #
            # We may have an __integration_config_id. This will only be
            # present if running Review Board < 5.0.3 with an upgraded
            # rbintegrations >= 3.1. If so, we can try to filter it.
            matching_configs = self.get_matching_configs(review_request)
            config_id = status_update.extra_data.get('__integration_config_id')

            if config_id is not None:
                matching_configs = [
                    _config
                    for _config in matching_configs
                    if _config.pk == config_id
                ]

            if not matching_configs:
                return
        else:
            # This is a StatusUpdate created with Review Board >= 5.0.3.
            matching_configs = [config]

        # If there's a change description associated with the status
        # update, then use the diff from that. Otherwise, choose the first
        # diffset on the review request.
        diffset: (DiffSet | None) = None

        try:
            if changedesc:
                if 'diff' in changedesc.fields_changed:
                    new_diff = changedesc.fields_changed['diff']['added'][0]
                    diffset = DiffSet.objects.get(pk=new_diff[2])
            else:
                diffset = (
                    DiffSet.objects
                    .filter(history=review_request.diffset_history_id)
                    .earliest()
                )
        except DiffSet.DoesNotExist:
            logger.error('Unable to determine diffset when running '
                         '%s for status update %d',
                         self.name, status_update.pk)
            return

        if diffset is None:
            return

        prep_data = BuildPrepData(
            changedesc=changedesc,
            configs=matching_configs,
            diffset=diffset,
            review_request=review_request,
            user=status_update.user)

        if not self.prepare_builds(prep_data):
            return

        if len(prep_data.configs) != 1:
            logger.error('Unable to determine the right configuration when '
                         'running %s for status update %d (%d configurations '
                         'found).',
                         self.name, status_update.pk, len(prep_data.configs))
            return

        config = prep_data.configs[0]

        self.set_starting(status_update)
        self._run_start_build(prep_data=prep_data,
                              config=config,
                              status_update=status_update)

    def _run_start_build(
        self,
        *,
        prep_data: BuildPrepData,
        config: IntegrationConfig,
        status_update: StatusUpdate,
    ) -> None:
        """Run start_build, handling exceptions.

        This wraps the subclass's :py:meth:`start_build`, catching any
        exceptions and updating the
        :py:class:`~reviewboard.reviews.models.status_update.StatusUpdate`
        accordingly.

        Args:
            prep_data (BuildPrepData):
                The builds preparation data containing information for the
                build.

            config (reviewboard.integrations.models.IntegrationConfig):
                The configuration for the integration triggering this build.

            status_update (reviewboard.reviews.models.status_update.
                           StatusUpdate):
                The status update indicating the status of this build.
        """
        try:
            self.start_build(config=config,
                             prep_data=prep_data,
                             status_update=status_update)
        except CIBuildError as e:
            self.set_error(status_update,
                           description=str(e),
                           url=e.url,
                           url_text=e.url_text)
        except Exception as e:
            logger.exception('Unexpected error running %s build for review '
                             'request %s and configuration %s: %s',
                             self.name, prep_data.review_request.pk,
                             config.pk, e)
            self.set_error(status_update,
                           description=f'internal error: {e}')



"""Pages for I Done This integration."""

from django.utils.translation import gettext_lazy as _
from reviewboard.accounts.pages import AccountPage

from rbintegrations.idonethis.forms import IDoneThisIntegrationAccountPageForm


class IDoneThisIntegrationAccountPage(AccountPage):
    """User account page for I Done This."""

    page_id = 'idonethis_account_page'
    page_title = _('I Done This')
    form_classes = [IDoneThisIntegrationAccountPageForm]



"""Unit tests for the Jenkins CI integration."""

from __future__ import annotations

import json
from urllib.error import HTTPError
from urllib.parse import urlencode
from typing import Optional, TYPE_CHECKING

import kgb
import reviewboard
from django.utils.encoding import force_bytes
from djblets.conditions import ConditionSet, Condition
from reviewboard.reviews.conditions import ReviewRequestRepositoriesChoice
from reviewboard.reviews.signals import status_update_request_run

from rbintegrations.jenkinsci.api import JenkinsAPI
from rbintegrations.jenkinsci.integration import JenkinsCIIntegration
from rbintegrations.testing.testcases import IntegrationTestCase

if TYPE_CHECKING:
    from reviewboard.reviews.models import ReviewRequest


class JenkinsCIIntegrationTests(IntegrationTestCase):
    """Tests for Jenkins CI."""

    integration_cls = JenkinsCIIntegration
    fixtures = ['test_scmtools', 'test_site', 'test_users']

    def test_build_new_review_request(self):
        """Testing JenkinsCIIntegration builds a new review request"""
        review_request = self._setup_build_requests()
        review_request.publish(review_request.submitter)

        self._check_build_requests(payload={
            'parameter': [
                {
                    'name': 'REVIEWBOARD_SERVER',
                    'value': 'http://example.com/'
                },
                {
                    'name': 'REVIEWBOARD_REVIEW_ID',
                    'value': review_request.display_id
                },
                {
                    'name': 'REVIEWBOARD_REVIEW_BRANCH',
                    'value': review_request.branch
                },
                {
                    'name': 'REVIEWBOARD_DIFF_REVISION',
                    'value': 1
                },
                {
                    'name': 'REVIEWBOARD_STATUS_UPDATE_ID',
                    'value': 1
                }
            ]
        })

    def test_build_new_review_request_with_local_site(self):
        """Testing JenkinsCIIntegration builds a new review request with a
        local site
        """
        review_request = self._setup_build_requests(with_local_site=True)
        review_request.publish(review_request.submitter)

        self._check_build_requests(payload={
            'parameter': [
                {
                    'name': 'REVIEWBOARD_SERVER',
                    'value': ('http://example.com/s/%s/' %
                              self.local_site_name)
                },
                {
                    'name': 'REVIEWBOARD_REVIEW_ID',
                    'value': review_request.display_id
                },
                {
                    'name': 'REVIEWBOARD_REVIEW_BRANCH',
                    'value': review_request.branch
                },
                {
                    'name': 'REVIEWBOARD_DIFF_REVISION',
                    'value': 1
                },
                {
                    'name': 'REVIEWBOARD_STATUS_UPDATE_ID',
                    'value': 1
                }
            ]
        })

    def test_manual_run_no_build_on_publish(self):
        """Testing lack of JenkinsCIIntegration build when a new review
        request is made with the run manually configuration
        """
        review_request = self._setup_build_requests(run_manually=True)
        review_request.publish(review_request.submitter)

        self._check_build_requests(expect_fetch_csrf_token=False,
                                   expect_open_request=False)

    def test_build_manual_run(self):
        """Testing JenkinsCIIntegration build via a manual trigger"""
        review_request = self._setup_build_requests(run_manually=True)

        status_update = \
            self.create_status_update(service_id='jenkins-ci',
                                      review_request=review_request)

        status_update_request_run.send(sender=self.__class__,
                                       status_update=status_update)

        self._check_build_requests(payload={
            'parameter': [
                {
                    'name': 'REVIEWBOARD_SERVER',
                    'value': 'http://example.com/',
                },
                {
                    'name': 'REVIEWBOARD_REVIEW_ID',
                    'value': review_request.display_id,
                },
                {
                    'name': 'REVIEWBOARD_REVIEW_BRANCH',
                    'value': review_request.branch
                },
                {
                    'name': 'REVIEWBOARD_DIFF_REVISION',
                    'value': 1,
                },
                {
                    'name': 'REVIEWBOARD_STATUS_UPDATE_ID',
                    'value': 1,
                }
            ]
        })

    def test_build_new_review_request_no_csrf_protection(self):
        """Testing that JenkinsCIIntegration builds a new review request
        without csrf protection
        """
        review_request = self._setup_build_requests(
            csrf_error=HTTPError('', 404, 'Not found', None, None))
        review_request.publish(review_request.submitter)

        self._check_build_requests(
            payload={
                'parameter': [
                    {
                        'name': 'REVIEWBOARD_SERVER',
                        'value': 'http://example.com/',
                    },
                    {
                        'name': 'REVIEWBOARD_REVIEW_ID',
                        'value': review_request.display_id,
                    },
                    {
                        'name': 'REVIEWBOARD_REVIEW_BRANCH',
                        'value': review_request.branch
                    },
                    {
                        'name': 'REVIEWBOARD_DIFF_REVISION',
                        'value': 1,
                    },
                    {
                        'name': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': 1,
                    }
                ],
            },
            expect_csrf_protection=False)

    def test_build_new_review_request_crumb_fetch_error(self):
        """Testing that JenkinsCIIntegration does not build when fetching the
        csrf token (or crumb) results in a non-404 error code
        """
        review_request = self._setup_build_requests(
            csrf_error=HTTPError('', 400, 'Bad request', None, None))
        review_request.publish(review_request.submitter)

        self._check_build_requests(expect_open_request=False)

    def test_job_name_variables(self) -> None:
        """Testing that JenkinsCIIntegration correctly replaces variables in
        a job name
        """
        review_request = self._setup_build_requests(
            job_name='{repository_name}/{branch_name}/1',
            branch='my/branch')
        review_request.publish(review_request.submitter)

        self._check_build_requests(
            url=(
                'http://localhost:8000/job/'
                'Test%20Repo/Name/my/branch/1/build'
            ),
            payload={
                'parameter': [
                    {
                        'name': 'REVIEWBOARD_SERVER',
                        'value': 'http://example.com/',
                    },
                    {
                        'name': 'REVIEWBOARD_REVIEW_ID',
                        'value': review_request.display_id,
                    },
                    {
                        'name': 'REVIEWBOARD_REVIEW_BRANCH',
                        'value': review_request.branch,
                    },
                    {
                        'name': 'REVIEWBOARD_DIFF_REVISION',
                        'value': 1,
                    },
                    {
                        'name': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': 1,
                    }
                ]
            })

    def test_job_name_variables_replaced_noslash(self) -> None:
        """Testing that JenkinsCIIntegration correctly replaces
        noslash-variant variables in a job name
        """
        review_request = self._setup_build_requests(
            job_name='{noslash_repository_name}/{noslash_branch_name}/1',
            branch='my/branch')
        review_request.publish(review_request.submitter)

        self._check_build_requests(
            url=(
                'http://localhost:8000/job/'
                'Test%20Repo_Name/my_branch/1/build'
            ),
            payload={
                'parameter': [
                    {
                        'name': 'REVIEWBOARD_SERVER',
                        'value': 'http://example.com/',
                    },
                    {
                        'name': 'REVIEWBOARD_REVIEW_ID',
                        'value': review_request.display_id,
                    },
                    {
                        'name': 'REVIEWBOARD_REVIEW_BRANCH',
                        'value': review_request.branch,
                    },
                    {
                        'name': 'REVIEWBOARD_DIFF_REVISION',
                        'value': 1,
                    },
                    {
                        'name': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': 1,
                    }
                ]
            })

    def test_job_name_variables_replaced_legacy(self) -> None:
        """Testing that JenkinsCIIntegration correctly replaces legacy
        variables in a job name
        """
        review_request = self._setup_build_requests(
            job_name='{repository}/{branch}/1',
            branch='my/branch')
        review_request.publish(review_request.submitter)

        self._check_build_requests(
            url=(
                'http://localhost:8000/job/'
                'Test%20Repo_Name/my_branch/1/build'
            ),
            payload={
                'parameter': [
                    {
                        'name': 'REVIEWBOARD_SERVER',
                        'value': 'http://example.com/',
                    },
                    {
                        'name': 'REVIEWBOARD_REVIEW_ID',
                        'value': review_request.display_id,
                    },
                    {
                        'name': 'REVIEWBOARD_REVIEW_BRANCH',
                        'value': review_request.branch,
                    },
                    {
                        'name': 'REVIEWBOARD_DIFF_REVISION',
                        'value': 1,
                    },
                    {
                        'name': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': 1,
                    }
                ]
            })

    def _create_config(self, job_name=None, with_local_site=False,
                       run_manually=False):
        """Create an integration config.

        Args:
            job_name (string, optional):
                The Jenkins job name, which is used in constructing the
                URL to start a build on Jenkins.

            with_local_site (bool, optional):
                Whether to create the configuration for a Local Site.

            run_manually (bool, optional):
                Whether to run JenkinsCIIntegration manually.

        Returns:
            reviewboard.integrations.models.IntegrationConfig:
            The resulting integration configuration.
        """
        choice = ReviewRequestRepositoriesChoice()

        condition_set = ConditionSet(conditions=[
            Condition(choice=choice,
                      operator=choice.get_operator('any'))
        ])

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        config = self.integration.create_config(name='Config 1',
                                                enabled=True,
                                                local_site=local_site)
        config.set('conditions', condition_set.serialize())
        config.set('jenkins_job_name', job_name or 'job_1')
        config.set('jenkins_endpoint', 'http://localhost:8000')
        config.set('jenkins_username', 'admin')
        config.set('jenkins_password', 'admin')
        config.set('run_manually', run_manually)
        config.save()

        return config

    def _setup_build_requests(
        self,
        *,
        with_local_site: bool = False,
        run_manually: bool = False,
        job_name: Optional[str] = None,
        branch: Optional[str] = 'my-branch',
        csrf_error: Optional[Exception] = None,
    ) -> ReviewRequest:
        """Set up state for build-related tests.

        This will set up a repository, review request, integration
        configuration, and spies for a build-related unit test.

        Args:
            with_local_site (bool, optional):
                Whether the tests should be performed against a Local Site.

            run_manually (bool, optional):
                Whether to create the integration configuration with
                ``run_manually`` set.

            job_name (str, optional):
                An explicit name or pattern to give the job.

            branch (str, optional):
                An optional branch name for the review request.

            csrf_error (Exception, optional):
                An optional exception to raise for the CSRF request.

                If not provided, a successful payload will be returned
                instead.

        Returns:
            reviewboard.reviews.models.ReviewRequest:
            The created review request.
        """
        repository = self.create_repository(
            name='Test Repo/Name',
            with_local_site=with_local_site)
        review_request = self.create_review_request(
            repository=repository,
            with_local_site=with_local_site,
            branch=branch)

        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save(update_fields=('base_commit_id',))

        self._create_config(with_local_site=with_local_site,
                            run_manually=run_manually,
                            job_name=job_name)
        self.integration.enable_integration()

        # We'll always perform a CSRF check before any build request. This
        # code works under that assumption. Assertions later will confirm
        # them.
        def _handle_csrf(*args, **kwarga):
            if csrf_error:
                raise csrf_error

            return json.dumps({
                'crumb': 'crumb123',
                'crumbRequestField': 'crumbField',
            }).encode('utf-8')

        self.spy_on(
            JenkinsAPI._open_request,
            owner=JenkinsAPI,
            op=kgb.SpyOpMatchInOrder([
                {
                    'call_fake': _handle_csrf,
                },
                {
                    'call_fake': lambda *args, **kwargs: b'',
                },
            ]))

        return review_request

    def _check_build_requests(self,
                              url='http://localhost:8000/job/job_1/build',
                              payload=None,
                              expect_fetch_csrf_token=True,
                              expect_open_request=True,
                              expect_csrf_protection=True):
        """Check the requests made via the API.

        Args:
            url (unicode, optional):
                The expected URL.

                This will be ignored if ``expect_open_request`` is ``False``.

            payload (dict):
                The expected JSON payload.

                This will be ignored if ``expect_open_request`` is ``False``.

            expect_fetch_csrf_token (bool, optional):
                Whether to expect :py:meth:`rbintegrations.jenkinsci.api.
                JenkinsAPI._fetch_csrf_token` to be called.

            expect_open_request (bool, optional):
                Whether to expect :py:meth:`rbintegrations.jenkinsci.api.
                JenkinsAPI._open_request` to be called.

            expect_csrf_protection (bool, optional):
                Whether to expect that CSRF protection was enabled for
                build requests.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        if expect_fetch_csrf_token and expect_open_request:
            self.assertSpyCallCount(JenkinsAPI._open_request, 2)
        elif expect_fetch_csrf_token or expect_open_request:
            self.assertSpyCallCount(JenkinsAPI._open_request, 1)
        else:
            self.assertSpyNotCalled(JenkinsAPI._open_request)

        call_index = 0

        # Check the CSRF call.
        if expect_fetch_csrf_token:
            self._check_http_request(
                call_index=call_index,
                url='http://localhost:8000/crumbIssuer/api/json')
            call_index += 1

        # Check the build request.
        if expect_open_request:
            if expect_csrf_protection:
                crumb = 'crumb123'
            else:
                crumb = None

            self._check_http_request(
                call_index=call_index,
                url=url,
                content_type='application/x-www-form-urlencoded',
                crumb=crumb,
                data=force_bytes(urlencode({
                    'json': json.dumps(payload, sort_keys=True),
                })))

    def _check_http_request(self, call_index, url, data=None,
                            content_type=None, crumb=None):
        """Check the contents of an HTTP request.

        This will ensure the request has the expected URL, payload, and
        headers.

        Args:
            call_index (int):
                The index of this particular HTTP call.

            url (unicode):
                The expected URL.

            data (bytes, optional):
                The expected request payload contents.

            content_type (unicode, optional):
                The expected value of the :mimeheader:`Content-type` header.

            crumb (unicode, optional):
                The expected value of the crumb header from a CSRF
                request.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        request = JenkinsAPI._open_request.calls[call_index].args[0]

        if reviewboard.VERSION[0] >= 4:
            request_url = request.url
        else:
            request_url = request.get_full_url()

        self.assertEqual(request_url, url)
        self.assertEqual(request.data, data)
        self.assertEqual(request.headers.get('Content-type'), content_type)
        self.assertEqual(request.headers.get('Crumbfield'), crumb)



"""Views for the Asana integration."""

from __future__ import annotations

import json
import logging
from typing import Iterable, cast

import asana
import asana.rest
from django.http import HttpRequest, HttpResponse
from django.utils.translation import gettext as _
from django.views.generic import View
from djblets.log import log_timed
from djblets.util.typing import JSONValue
from reviewboard.accounts.mixins import CheckLoginRequiredViewMixin
from reviewboard.integrations.base import get_integration_manager
from reviewboard.reviews.views import ReviewRequestViewMixin
from reviewboard.site.mixins import CheckLocalSiteAccessViewMixin


logger = logging.getLogger(__name__)


class AsanaTaskSearchView(ReviewRequestViewMixin, View):
    """The view to search for tasks (for use with auto-complete)."""

    def get(
        self,
        request: HttpRequest,
        **kwargs,
    ) -> HttpResponse:
        """Perform a search for tasks.

        Args:
            request (django.http.HttpRequest):
                The request.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.http.HttpResponse:
            A response containing JSON with a list of tasks matching the
            search criteria.
        """
        from rbintegrations.asana.integration import AsanaIntegration
        integration_id = AsanaIntegration.integration_id
        assert integration_id is not None

        integration_manager = get_integration_manager()
        integration = cast(
            AsanaIntegration,
            integration_manager.get_integration(integration_id))

        configs = (
            config
            for config in integration.get_configs(
                self.review_request.local_site)
            if config.match_conditions(form_cls=integration.config_form_cls,
                                       review_request=self.review_request)
        )

        results = []

        params = {
            'type': 'task',
            'query': request.GET.get('q'),
            'count': 20,
            'opt_fields': 'completed,name,notes',
        }

        for config in configs:
            workspace = config.settings.get('asana_workspace')

            with log_timed(f'Requesting tasks from Asana for workspace '
                           f'{workspace}',
                           logger=logger) as log_timer:
                try:
                    if not workspace:
                        raise ValueError(_(
                            'The workspace ID is not configured.'
                        ))

                    asana_config = asana.Configuration()
                    asana_config.access_token = \
                        config.settings['asana_access_token']

                    api_client = asana.ApiClient(asana_config)
                    typeahead_api = asana.TypeaheadApi(api_client)

                    workspace = config.settings['asana_workspace']

                    tasks = cast(
                        Iterable[JSONValue],
                        typeahead_api.typeahead_for_workspace(
                            workspace, 'task', params))

                    results.append({
                        'workspace': config.settings['asana_workspace_name'],
                        'workspace_id': workspace,
                        'tasks': list(tasks),
                    })
                except Exception as e:
                    logger.exception('[%s] Unexpected error when searching '
                                     'for Asana tasks: %s',
                                     log_timer.trace_id, e)

        return HttpResponse(json.dumps(results),
                            content_type='application/json')


class AsanaWorkspaceListView(CheckLoginRequiredViewMixin,
                             CheckLocalSiteAccessViewMixin, View):
    """The view to fetch the available workspaces.

    This is used by the integration config form to select a workspace.
    """

    def get(
        self,
        request: HttpRequest,
    ) -> HttpResponse:
        """Return the available workspaces.

        Args:
            request (django.http.HttpRequest):
                The request.

        Returns:
            django.http.HttpResponse:
            A response containing JSON with a list of available Asana
            workspaces.
        """
        asana_api_key = request.GET.get('api_key')
        assert asana_api_key is not None

        try:
            configuration = asana.Configuration()
            configuration.access_token = asana_api_key

            api_client = asana.ApiClient(configuration)
            workspaces_api = asana.WorkspacesApi(api_client)
            workspaces = cast(
                Iterable[JSONValue],
                workspaces_api.get_workspaces({}))

            results = {
                'result': 'success',
                'data': list(workspaces),
            }
        except asana.rest.ApiException as e:
            results = {
                'result': 'error',
                'error': f'Could not communicate with Asana: {e.reason}',
            }
        except Exception as e:
            results = {
                'result': 'error',
                'error': str(e),
            }

        return HttpResponse(json.dumps(results),
                            content_type='application/json')



"""Unit tests for rbintegrations.baseci.BaseCIIntegration.

Version Added:
    3.1
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional, TYPE_CHECKING

import kgb
from django.db.models import Q
from djblets.conditions import ConditionSet, Condition
from djblets.testing.decorators import add_fixtures
from kgb.ops import BaseSpyOperation
from reviewboard.diffviewer.models import DiffSet
from reviewboard.integrations.base import get_integration_manager
from reviewboard.integrations.forms import IntegrationConfigForm
from reviewboard.integrations.models import IntegrationConfig
from reviewboard.reviews.conditions import ReviewRequestRepositoriesChoice
from reviewboard.reviews.models import StatusUpdate
from reviewboard.site.models import LocalSite
from typing_extensions import TypedDict

from rbintegrations.baseci.errors import CIBuildError
from rbintegrations.baseci.integration import BaseCIIntegration, BuildPrepData
from rbintegrations.testing.testcases import IntegrationTestCase
from rbintegrations.travisci.integration import TravisCIIntegration
from rbintegrations.util.conditions import ReviewRequestConditionsField

if TYPE_CHECKING:
    from reviewboard.reviews.models import ReviewRequest


class SetupRunTestResults(TypedDict):
    """Results from setting up a run test for a build integration.

    Version Added:
        3.1
    """

    #: The configuration to use for the build.
    config: Optional[IntegrationConfig]

    #: The DiffSets created for the test.
    diffsets: List[DiffSet]

    #: The review request created for the test.
    review_request: ReviewRequest


class MyCIIntegrationConfigForm(IntegrationConfigForm):
    """Test form for BaseCIIntegration tests.

    Version Added:
        3.1
    """

    conditions = ReviewRequestConditionsField()


class MyCIIntegration(BaseCIIntegration):
    """Test integration for BaseCIIntegration tests.

    Version Added:
        3.1
    """

    name = 'My CI'
    integration_id = 'my-ci'
    config_form_cls = MyCIIntegrationConfigForm
    status_update_service_id = 'my-ci'
    bot_username = 'test-ci-user'
    bot_user_first_name = 'Test'
    bot_user_last_name = 'CI'

    def prepare_builds(
        self,
        prep_data: BuildPrepData,
    ) -> bool:
        prep_data.extra_state['my_state'] = 123

        return True

    def start_build(
        self,
        *,
        prep_data: BuildPrepData,
        config: IntegrationConfig,
        status_update: StatusUpdate,
    ) -> None:
        pass


class BaseCIIntegrationTests(IntegrationTestCase):
    """Unit tests for BaseCIIntegration.

    Version Added:
        3.1
    """

    integration_cls = MyCIIntegration
    fixtures = ['test_users']

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the test case class."""
        super().setUpClass()

        get_integration_manager().register_integration_class(
            cls.integration_cls)

    @classmethod
    def tearDownClass(cls) -> None:
        """Tear down the test case class."""
        get_integration_manager().unregister_integration_class(
            cls.integration_cls)

        super().tearDownClass()

    def test_get_or_create_api_token_with_new(self) -> None:
        """Testing BaseCIIntegration.get_or_create_api_token with new token"""
        user = self.create_user()

        with self.assertNumQueries(2):
            token = self.integration.get_or_create_api_token(
                user=user,
                local_site=None)

        self.assertIsNotNone(token.pk)
        self.assertEqual(token.user, user)
        self.assertIsNone(token.local_site)

    def test_get_or_create_api_token_with_new_and_local_site(self) -> None:
        """Testing BaseCIIntegration.get_or_create_api_token with new token
        on LocalSite
        """
        user = self.create_user()
        local_site = self.create_local_site()

        with self.assertNumQueries(2):
            token = self.integration.get_or_create_api_token(
                user=user,
                local_site=local_site)

        self.assertIsNotNone(token.pk)
        self.assertEqual(token.user, user)
        self.assertEqual(token.local_site, local_site)

    def test_get_or_create_api_token_with_existing(self) -> None:
        """Testing BaseCIIntegration.get_or_create_api_token with existing
        token
        """
        user = self.create_user()

        with self.assertNumQueries(2):
            token1 = self.integration.get_or_create_api_token(
                user=user,
                local_site=None)

        with self.assertNumQueries(1):
            token2 = self.integration.get_or_create_api_token(
                user=user,
                local_site=None)

        self.assertEqual(token1, token2)

    def test_get_or_create_api_token_with_existing_and_local_site(
        self,
    ) -> None:
        """Testing BaseCIIntegration.get_or_create_api_token with existing
        token on LocalSite
        """
        user = self.create_user()
        local_site = self.create_local_site()

        with self.assertNumQueries(2):
            token1 = self.integration.get_or_create_api_token(
                user=user,
                local_site=local_site)

        with self.assertNumQueries(1):
            token2 = self.integration.get_or_create_api_token(
                user=user,
                local_site=local_site)

        self.assertEqual(token1, token2)

    def test_get_or_create_user_with_new(self) -> None:
        """Testing BaseCIIntegration.get_or_create_user with new user"""
        user = self.integration.get_or_create_user()

        self.assertIsNotNone(user.pk)
        self.assertAttrsEqual(
            user,
            {
                'email': 'noreply@example.com',
                'first_name': 'Test',
                'last_name': 'CI',
                'username': 'test-ci-user',
            })
        self.assertFalse(user.get_profile().should_send_email)

    def test_get_or_create_user_with_existing(self) -> None:
        """Testing BaseCIIntegration.get_or_create_user with existing user"""
        integration = self.integration
        user1 = integration.get_or_create_user()
        user2 = integration.get_or_create_user()

        self.assertEqual(user1, user2)

    @add_fixtures(['test_scmtools'])
    def test_get_matching_configs(self) -> None:
        """Testing BaseCIIntegration.get_matching_configs"""
        local_site = self.create_local_site()

        # These will match.
        config1 = self._create_config(name='config1')
        config2 = self._create_config(name='config2')

        # These will not match.
        self._create_config(name='config3',
                            integration_id=TravisCIIntegration.integration_id)
        self._create_config(name='config4',
                            enabled=False)
        self._create_config(name='config5',
                            with_condition_match=False)
        self._create_config(name='config6',
                            local_site=local_site)

        user = self.create_user()
        review_request = self.create_review_request(submitter=user,
                                                    create_repository=True)

        self.assertEqual(
            self.integration.get_matching_configs(
                review_request=review_request),
            [config1, config2])

    @add_fixtures(['test_scmtools', 'test_site'])
    def test_get_matching_configs_and_local_site(self) -> None:
        """Testing BaseCIIntegration.get_matching_configs and Local Site"""
        local_site1 = self.create_local_site(name='test-site-1')
        local_site2 = self.create_local_site(name='test-site-2')

        # These will match.
        config1 = self._create_config(name='config1',
                                      local_site=local_site1)
        config2 = self._create_config(name='config2',
                                      local_site=local_site1)

        # These will not match.
        self._create_config(name='config3',
                            integration_id=TravisCIIntegration.integration_id,
                            local_site=local_site1)
        self._create_config(name='config4',
                            enabled=False,
                            local_site=local_site1)
        self._create_config(name='config5',
                            with_condition_match=False,
                            local_site=local_site1)
        self._create_config(name='config6')
        self._create_config(name='config7',
                            local_site=local_site2)

        user = self.create_user()
        review_request = self.create_review_request(submitter=user,
                                                    create_repository=True,
                                                    local_site=local_site1)

        self.assertEqual(
            self.integration.get_matching_configs(
                review_request=review_request),
            [config1, config2])

    def test_update_status_with_no_new_data(self) -> None:
        """Testing BaseCIIntegration.update_status with no new data"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(review_request)
        old_timestamp = status_update.timestamp

        with self.assertNumQueries(0):
            self.integration.update_status(status_update)

        self.assertEqual(status_update.timestamp, old_timestamp)
        self.assertIsNone(status_update.timeout)

    def test_update_status_with_noops(self) -> None:
        """Testing BaseCIIntegration.update_status with no no-op changes"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(
            review_request,
            state=StatusUpdate.PENDING,
            description='My description.',
            url='https://example.com',
            url_text='Link')
        old_timestamp = status_update.timestamp

        with self.assertNumQueries(0):
            self.integration.update_status(
                status_update,
                state=StatusUpdate.PENDING,
                description='My description.',
                url='https://example.com',
                url_text='Link')

        self.assertEqual(status_update.timestamp, old_timestamp)
        self.assertIsNone(status_update.timeout)

    def test_update_status_with_new_data(self) -> None:
        """Testing BaseCIIntegration.update_status with new data"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(
            review_request,
            state=StatusUpdate.PENDING,
            description='description.',
            url='https://example.com',
            url_text='Link')
        old_timestamp = status_update.timestamp

        queries = [
            {
                'model': StatusUpdate,
                'type': 'UPDATE',
                'where': Q(pk=status_update.pk),
            },
        ]

        with self.assertQueries(queries):
            self.integration.update_status(
                status_update,
                state=StatusUpdate.DONE_SUCCESS,
                description='new description.',
                url='https://example.com/new/',
                url_text='New link')

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'new description.',
                'state': StatusUpdate.DONE_SUCCESS,
                'timeout': None,
                'url': 'https://example.com/new/',
                'url_text': 'New link',
            })
        self.assertGreater(status_update.timestamp, old_timestamp)

        # Make sure this all saved.
        status_update.refresh_from_db()

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'new description.',
                'state': StatusUpdate.DONE_SUCCESS,
                'url': 'https://example.com/new/',
                'url_text': 'New link',
            })
        self.assertGreater(status_update.timestamp, old_timestamp)

    def test_update_status_with_save_false(self) -> None:
        """Testing BaseCIIntegration.update_status with save=False"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(
            review_request,
            state=StatusUpdate.PENDING,
            description='description.',
            url='https://example.com',
            url_text='Link')
        old_timestamp = status_update.timestamp

        with self.assertNumQueries(0):
            self.integration.update_status(
                status_update,
                state=StatusUpdate.DONE_SUCCESS,
                description='new description.',
                url='https://example.com/new/',
                url_text='New link',
                save=False)

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'new description.',
                'state': StatusUpdate.DONE_SUCCESS,
                'timeout': None,
                'url': 'https://example.com/new/',
                'url_text': 'New link',
            })
        self.assertGreater(status_update.timestamp, old_timestamp)

        # Make sure none of that saved.
        status_update.refresh_from_db()

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'description.',
                'state': StatusUpdate.PENDING,
                'timestamp': old_timestamp,
                'timeout': None,
                'url': 'https://example.com',
                'url_text': 'Link',
            })

    def test_set_waiting(self) -> None:
        """Testing BaseCIIntegration.set_waiting"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(
            review_request,
            state=StatusUpdate.PENDING,
            description='My description.',
            url='https://example.com',
            url_text='Link')
        old_timestamp = status_update.timestamp

        self.integration.set_waiting(status_update)

        attrs = {
            'description': 'waiting to run.',
            'state': StatusUpdate.NOT_YET_RUN,
            'timeout': None,
            'url': 'https://example.com',
            'url_text': 'Link',
        }

        self.assertAttrsEqual(status_update, attrs)
        self.assertGreater(status_update.timestamp, old_timestamp)

        # Make sure this all saved.
        status_update.refresh_from_db()

        self.assertAttrsEqual(status_update, attrs)
        self.assertGreater(status_update.timestamp, old_timestamp)

    def test_set_waiting_with_state(self) -> None:
        """Testing BaseCIIntegration.set_waiting with new custom state"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(
            review_request,
            state=StatusUpdate.PENDING,
            description='description.',
            url='https://example.com',
            url_text='Link')
        old_timestamp = status_update.timestamp

        self.integration.set_waiting(
            status_update,
            description='new description.',
            url='https://example.com/new/',
            url_text='New link')

        attrs = {
            'description': 'new description.',
            'state': StatusUpdate.NOT_YET_RUN,
            'timeout': None,
            'url': 'https://example.com/new/',
            'url_text': 'New link',
        }

        self.assertAttrsEqual(status_update, attrs)
        self.assertGreater(status_update.timestamp, old_timestamp)

        # Make sure this all saved.
        status_update.refresh_from_db()

        self.assertAttrsEqual(status_update, attrs)
        self.assertGreater(status_update.timestamp, old_timestamp)

    def test_set_waiting_with_save_false(self) -> None:
        """Testing BaseCIIntegration.set_waiting with save=False"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(
            review_request,
            state=StatusUpdate.PENDING,
            description='description.',
            url='https://example.com',
            url_text='Link')
        old_timestamp = status_update.timestamp

        self.integration.set_waiting(
            status_update,
            description='new description.',
            url='https://example.com/new/',
            url_text='New link',
            save=False)

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'new description.',
                'state': StatusUpdate.NOT_YET_RUN,
                'timeout': None,
                'url': 'https://example.com/new/',
                'url_text': 'New link',
            })
        self.assertGreater(status_update.timestamp, old_timestamp)

        # Make sure none of that saved.
        status_update.refresh_from_db()

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'description.',
                'state': StatusUpdate.PENDING,
                'timestamp': old_timestamp,
                'url': 'https://example.com',
                'url_text': 'Link',
            })

    def test_set_starting(self) -> None:
        """Testing BaseCIIntegration.set_starting"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(
            review_request,
            state=StatusUpdate.ERROR,
            description='My description.',
            url='https://example.com',
            url_text='Link')
        old_timestamp = status_update.timestamp

        self.integration.set_starting(status_update)

        attrs = {
            'description': 'starting build...',
            'state': StatusUpdate.PENDING,
            'timeout': None,
            'url': 'https://example.com',
            'url_text': 'Link',
        }

        self.assertAttrsEqual(status_update, attrs)
        self.assertGreater(status_update.timestamp, old_timestamp)

        # Make sure this all saved.
        status_update.refresh_from_db()

        self.assertAttrsEqual(status_update, attrs)
        self.assertGreater(status_update.timestamp, old_timestamp)

    def test_set_starting_with_state(self) -> None:
        """Testing BaseCIIntegration.set_starting with new custom state"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(
            review_request,
            state=StatusUpdate.ERROR,
            description='description.',
            url='https://example.com',
            url_text='Link')
        old_timestamp = status_update.timestamp

        self.integration.set_starting(
            status_update,
            description='new description.',
            url='https://example.com/new/',
            url_text='New link')

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'new description.',
                'state': StatusUpdate.PENDING,
                'timeout': None,
                'url': 'https://example.com/new/',
                'url_text': 'New link',
            })
        self.assertGreater(status_update.timestamp, old_timestamp)

        # Make sure this all saved.
        status_update.refresh_from_db()

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'new description.',
                'state': StatusUpdate.PENDING,
                'timeout': None,
                'url': 'https://example.com/new/',
                'url_text': 'New link',
            })
        self.assertGreater(status_update.timestamp, old_timestamp)

    def test_set_starting_with_save_false(self) -> None:
        """Testing BaseCIIntegration.set_starting with save=False"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(
            review_request,
            state=StatusUpdate.ERROR,
            description='description.',
            url='https://example.com',
            url_text='Link')
        old_timestamp = status_update.timestamp

        self.integration.set_starting(
            status_update,
            description='new description.',
            url='https://example.com/new/',
            url_text='New link',
            save=False)

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'new description.',
                'state': StatusUpdate.PENDING,
                'timeout': None,
                'url': 'https://example.com/new/',
                'url_text': 'New link',
            })
        self.assertGreater(status_update.timestamp, old_timestamp)

        # Make sure none of that saved.
        status_update.refresh_from_db()

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'description.',
                'state': StatusUpdate.ERROR,
                'timeout': None,
                'timestamp': old_timestamp,
                'url': 'https://example.com',
                'url_text': 'Link',
            })

    def test_set_error(self) -> None:
        """Testing BaseCIIntegration.set_error"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(
            review_request,
            state=StatusUpdate.PENDING,
            description='My description.',
            url='https://example.com',
            url_text='Link')
        old_timestamp = status_update.timestamp

        self.integration.set_error(status_update)

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'internal error.',
                'state': StatusUpdate.ERROR,
                'timeout': None,
                'url': 'https://example.com',
                'url_text': 'Link',
            })
        self.assertGreater(status_update.timestamp, old_timestamp)

        # Make sure this all saved.
        status_update.refresh_from_db()

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'internal error.',
                'state': StatusUpdate.ERROR,
                'timeout': None,
                'url': 'https://example.com',
                'url_text': 'Link',
            })
        self.assertGreater(status_update.timestamp, old_timestamp)

    def test_set_error_with_state(self) -> None:
        """Testing BaseCIIntegration.set_error with new custom state"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(
            review_request,
            state=StatusUpdate.PENDING,
            description='description.',
            url='https://example.com',
            url_text='Link')
        old_timestamp = status_update.timestamp

        self.integration.set_error(
            status_update,
            description='new description.',
            url='https://example.com/new/',
            url_text='New link')

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'new description.',
                'state': StatusUpdate.ERROR,
                'timeout': None,
                'url': 'https://example.com/new/',
                'url_text': 'New link',
            })
        self.assertGreater(status_update.timestamp, old_timestamp)

        # Make sure this all saved.
        status_update.refresh_from_db()

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'new description.',
                'state': StatusUpdate.ERROR,
                'timeout': None,
                'url': 'https://example.com/new/',
                'url_text': 'New link',
            })
        self.assertGreater(status_update.timestamp, old_timestamp)

    def test_set_error_with_save_false(self) -> None:
        """Testing BaseCIIntegration.set_error with save=False"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(
            review_request,
            state=StatusUpdate.PENDING,
            description='description.',
            url='https://example.com',
            url_text='Link')
        old_timestamp = status_update.timestamp

        self.integration.set_error(
            status_update,
            description='new description.',
            url='https://example.com/new/',
            url_text='New link',
            save=False)

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'new description.',
                'state': StatusUpdate.ERROR,
                'timeout': None,
                'url': 'https://example.com/new/',
                'url_text': 'New link',
            })
        self.assertGreater(status_update.timestamp, old_timestamp)

        # Make sure none of that saved.
        status_update.refresh_from_db()

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'description.',
                'state': StatusUpdate.PENDING,
                'timeout': None,
                'timestamp': old_timestamp,
                'url': 'https://example.com',
                'url_text': 'Link',
            })

    @add_fixtures(['test_scmtools'])
    def test_on_publish(self) -> None:
        """Testing BaseCIIntegration._on_review_request_published"""
        integration = self.integration

        info = self._setup_run_test()
        config = info['config']
        diffsets = info['diffsets']
        review_request = info['review_request']

        assert config

        self.assertSpyCalled(integration.start_build)
        call_kwargs = integration.start_build.last_call.kwargs
        prep_data = call_kwargs['prep_data']
        status_update = call_kwargs['status_update']

        self.assertEqual(call_kwargs['config'], config)

        self.assertAttrsEqual(
            prep_data,
            {
                'changedesc': None,
                'configs': [config],
                'diffset': diffsets[0],
                'extra_state': {
                    'my_state': 123,
                },
                'review_request': review_request,
            })
        self.assertEqual(prep_data.user.username, 'test-ci-user')

        self.assertIsNotNone(status_update.pk)
        self.assertAttrsEqual(
            status_update,
            {
                'change_description': None,
                'description': 'starting build...',
                'extra_data': {
                    '__integration_config_id': config.pk,
                    'can_retry': True,
                },
                'review_request': review_request,
                'service_id': 'my-ci',
                'state': StatusUpdate.PENDING,
                'summary': 'My CI',
                'timeout': None,
                'user': prep_data.user,
            })

    @add_fixtures(['test_scmtools'])
    def test_on_publish_with_changedesc(self) -> None:
        """Testing BaseCIIntegration._on_review_request_published with
        ChangeDescription
        """
        integration = self.integration

        info = self._setup_run_test(with_changedesc=True)
        config = info['config']
        diffsets = info['diffsets']
        review_request = info['review_request']

        assert config

        self.assertSpyCalled(integration.start_build)
        call_kwargs = integration.start_build.last_call.kwargs
        build_config = call_kwargs['config']
        prep_data = call_kwargs['prep_data']
        status_update = call_kwargs['status_update']

        self.assertEqual(build_config, config)

        self.assertAttrsEqual(
            prep_data,
            {
                'configs': [config],
                'diffset': diffsets[1],
                'extra_state': {
                    'my_state': 123,
                },
                'review_request': review_request,
            })
        self.assertIsNotNone(prep_data.changedesc)
        self.assertEqual(prep_data.user.username, 'test-ci-user')

        self.assertAttrsEqual(
            status_update,
            {
                'change_description': prep_data.changedesc,
                'description': 'starting build...',
                'extra_data': {
                    '__integration_config_id': config.pk,
                    'can_retry': True,
                },
                'review_request': review_request,
                'service_id': 'my-ci',
                'state': StatusUpdate.PENDING,
                'summary': 'My CI',
                'timeout': None,
                'user': prep_data.user,
            })
        self.assertIsNotNone(status_update.pk)

    @add_fixtures(['test_scmtools'])
    def test_on_publish_with_build_error(self) -> None:
        """Testing BaseCIIntegration._on_review_request_published with
        CIBuildError
        """
        integration = self.integration

        self._setup_run_test(spy_op=kgb.SpyOpRaise(CIBuildError(
            'some error.',
            url='https://example.com/error/',
            url_text='Error')))

        self.assertSpyCalled(integration.start_build)
        call_kwargs = integration.start_build.last_call.kwargs
        status_update = call_kwargs['status_update']

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'some error.',
                'state': StatusUpdate.ERROR,
                'timeout': None,
                'url': 'https://example.com/error/',
                'url_text': 'Error',
            })

    @add_fixtures(['test_scmtools'])
    def test_on_publish_with_exception(self) -> None:
        """Testing BaseCIIntegration._on_review_request_published with
        exception
        """
        integration = self.integration

        self._setup_run_test(spy_op=kgb.SpyOpRaise(Exception('oh no.')))

        self.assertSpyCalled(integration.start_build)
        call_kwargs = integration.start_build.last_call.kwargs
        status_update = call_kwargs['status_update']

        self.assertAttrsEqual(
            status_update,
            {
                'description': 'internal error: oh no.',
                'state': StatusUpdate.ERROR,
                'timeout': None,
            })

    @add_fixtures(['test_scmtools'])
    def test_on_publish_without_matching_config(self) -> None:
        """Testing BaseCIIntegration._on_review_request_published without
        matching configuration
        """
        self._setup_run_test(with_config=False)

        self.assertSpyNotCalled(self.integration.start_build)

    @add_fixtures(['test_scmtools'])
    def test_on_publish_without_diffset(self) -> None:
        """Testing BaseCIIntegration._on_review_request_published without
        DiffSet
        """
        self._setup_run_test(with_diffset=False)

        self.assertSpyNotCalled(self.integration.start_build)

    @add_fixtures(['test_scmtools'])
    def test_on_publish_without_changedesc_diffset(self) -> None:
        """Testing BaseCIIntegration._on_review_request_published without
        new diff in review request update
        """
        self._setup_run_test(with_changedesc=True,
                             with_diffset=False)

        self.assertSpyNotCalled(self.integration.start_build)

    @add_fixtures(['test_scmtools'])
    def test_on_publish_with_prepare_builds_false(self) -> None:
        """Testing BaseCIIntegration._on_review_request_published with
        prepare_builds() returning False
        """
        integration = self.integration

        self.spy_on(integration.prepare_builds,
                    op=kgb.SpyOpReturn(False))

        self._setup_run_test()

        self.assertSpyNotCalled(integration.start_build)

    @add_fixtures(['test_scmtools'])
    def test_manual_run(self) -> None:
        """Testing BaseCIIntegration._on_status_update_request_run with
        manual run
        """
        integration = self.integration

        info = self._setup_run_test(with_manual_run=True)
        config = info['config']
        diffsets = info['diffsets']
        review_request = info['review_request']

        assert config

        self.assertSpyNotCalled(integration.start_build)

        # Check the status update created on publish.
        status_update = StatusUpdate.objects.get()
        user = integration.get_or_create_user()

        self.assertIsNotNone(status_update.pk)
        self.assertAttrsEqual(
            status_update,
            {
                'change_description': None,
                'description': 'waiting to run.',
                'extra_data': {
                    '__integration_config_id': config.pk,
                    'can_retry': True,
                },
                'review_request': review_request,
                'service_id': 'my-ci',
                'state': StatusUpdate.NOT_YET_RUN,
                'summary': 'My CI',
                'timeout': None,
                'user': user,
            })

        # Perform the manual run.
        status_update.run()

        self.assertSpyCalled(integration.start_build)
        call_kwargs = integration.start_build.last_call.kwargs
        prep_data = call_kwargs['prep_data']

        self.assertEqual(call_kwargs['status_update'], status_update)
        self.assertEqual(call_kwargs['config'], config)

        status_update = call_kwargs['status_update']

        self.assertIsNotNone(status_update.pk)
        self.assertAttrsEqual(
            status_update,
            {
                'change_description': None,
                'description': 'starting build...',
                'extra_data': {
                    '__integration_config_id': config.pk,
                    'can_retry': True,
                },
                'review_request': review_request,
                'service_id': 'my-ci',
                'state': StatusUpdate.PENDING,
                'summary': 'My CI',
                'timeout': None,
                'user': user,
            })

        self.assertAttrsEqual(
            prep_data,
            {
                'changedesc': None,
                'configs': [config],
                'diffset': diffsets[0],
                'extra_state': {
                    'my_state': 123,
                },
                'review_request': review_request,
            })
        self.assertEqual(prep_data.user.username, 'test-ci-user')

        self.assertIsNotNone(status_update.pk)
        self.assertAttrsEqual(
            status_update,
            {
                'change_description': None,
                'description': 'starting build...',
                'extra_data': {
                    '__integration_config_id': config.pk,
                    'can_retry': True,
                },
                'review_request': review_request,
                'service_id': 'my-ci',
                'state': StatusUpdate.PENDING,
                'summary': 'My CI',
                'timeout': None,
                'user': prep_data.user,
            })

    @add_fixtures(['test_scmtools'])
    def test_manual_run_with_changedesc(self) -> None:
        """Testing BaseCIIntegration._on_status_update_request_run with
        manual run on update to review request
        """
        integration = self.integration

        info = self._setup_run_test(with_changedesc=True,
                                    with_manual_run=True)
        config = info['config']
        diffsets = info['diffsets']
        review_request = info['review_request']

        assert config

        self.assertSpyNotCalled(integration.start_build)

        user = integration.get_or_create_user()

        # Check the status updates created on publish.
        status_updates = StatusUpdate.objects.all()
        self.assertEqual(len(status_updates), 2)

        status_update = status_updates[1]

        self.assertIsNotNone(status_update.pk)
        self.assertAttrsEqual(
            status_update,
            {
                'description': 'waiting to run.',
                'extra_data': {
                    '__integration_config_id': config.pk,
                    'can_retry': True,
                },
                'review_request': review_request,
                'service_id': 'my-ci',
                'state': StatusUpdate.NOT_YET_RUN,
                'summary': 'My CI',
                'timeout': None,
                'user': user,
            })
        self.assertIsNotNone(status_update.change_description)

        changedesc = status_update.change_description

        # Perform the manual run.
        status_update.run()

        self.assertSpyCalled(integration.start_build)
        call_kwargs = integration.start_build.last_call.kwargs
        prep_data = call_kwargs['prep_data']

        self.assertEqual(call_kwargs['status_update'], status_update)
        self.assertEqual(call_kwargs['config'], config)

        status_update = call_kwargs['status_update']

        self.assertAttrsEqual(
            prep_data,
            {
                'changedesc': changedesc,
                'configs': [config],
                'diffset': diffsets[1],
                'extra_state': {
                    'my_state': 123,
                },
                'review_request': review_request,
            })
        self.assertEqual(prep_data.user.username, 'test-ci-user')

        self.assertIsNotNone(status_update.pk)
        self.assertAttrsEqual(
            status_update,
            {
                'change_description': changedesc,
                'description': 'starting build...',
                'extra_data': {
                    '__integration_config_id': config.pk,
                    'can_retry': True,
                },
                'review_request': review_request,
                'service_id': 'my-ci',
                'state': StatusUpdate.PENDING,
                'summary': 'My CI',
                'timeout': None,
                'user': prep_data.user,
            })

    @add_fixtures(['test_scmtools'])
    def test_manual_run_with_multiple_configs(self) -> None:
        """Testing BaseCIIntegration._on_status_update_request_run with
        multiple matching configs in database
        """
        integration = self.integration
        integration_id = integration.integration_id

        config1 = self._create_config(name='bad-config-1',
                                      integration_id=integration_id,
                                      with_manual_run=True)

        info = self._setup_run_test(with_manual_run=True)
        config2 = info['config']
        diffsets = info['diffsets']
        review_request = info['review_request']

        assert config2

        self._create_config(name='bad-config-2',
                            integration_id=integration_id,
                            with_manual_run=True)

        self.assertSpyNotCalled(integration.start_build)

        # Check the status updates created on publish.
        #
        # There should be only 2, since we did a publish after config1 and
        # config2 were published but before config3.
        status_updates = list(StatusUpdate.objects.all())

        self.assertEqual(len(status_updates), 2)
        self.assertEqual(status_updates[0].integration_config, config1)
        self.assertEqual(status_updates[1].integration_config, config2)

        # We want to test against the middle status update (the "good" one).
        status_update = status_updates[1]

        user = integration.get_or_create_user()

        self.assertIsNotNone(status_update.pk)
        self.assertAttrsEqual(
            status_update,
            {
                'change_description': None,
                'description': 'waiting to run.',
                'extra_data': {
                    '__integration_config_id': config2.pk,
                    'can_retry': True,
                },
                'review_request': review_request,
                'service_id': 'my-ci',
                'state': StatusUpdate.NOT_YET_RUN,
                'summary': 'My CI',
                'timeout': None,
                'user': user,
            })

        # Perform the manual run.
        status_update.run()

        self.assertSpyCalled(integration.start_build)
        call_kwargs = integration.start_build.last_call.kwargs
        prep_data = call_kwargs['prep_data']

        self.assertEqual(call_kwargs['status_update'], status_update)
        self.assertEqual(call_kwargs['config'], config2)

        status_update = call_kwargs['status_update']

        self.assertIsNotNone(status_update.pk)
        self.assertAttrsEqual(
            status_update,
            {
                'change_description': None,
                'description': 'starting build...',
                'extra_data': {
                    '__integration_config_id': config2.pk,
                    'can_retry': True,
                },
                'review_request': review_request,
                'service_id': 'my-ci',
                'state': StatusUpdate.PENDING,
                'summary': 'My CI',
                'timeout': None,
                'user': user,
            })

        self.assertAttrsEqual(
            prep_data,
            {
                'changedesc': None,
                'configs': [config2],
                'diffset': diffsets[0],
                'extra_state': {
                    'my_state': 123,
                },
                'review_request': review_request,
            })
        self.assertEqual(prep_data.user.username, 'test-ci-user')

        self.assertIsNotNone(status_update.pk)
        self.assertAttrsEqual(
            status_update,
            {
                'change_description': None,
                'description': 'starting build...',
                'extra_data': {
                    '__integration_config_id': config2.pk,
                    'can_retry': True,
                },
                'review_request': review_request,
                'service_id': 'my-ci',
                'state': StatusUpdate.PENDING,
                'summary': 'My CI',
                'timeout': None,
                'user': prep_data.user,
            })

    @add_fixtures(['test_scmtools'])
    def test_manual_run_with_build_error(self) -> None:
        """Testing BaseCIIntegration._on_status_update_request_run with
        CIBuildError
        """
        integration = self.integration

        self._setup_run_test(
            with_manual_run=True,
            spy_op=kgb.SpyOpRaise(CIBuildError(
                'some error.',
                url='https://example.com/error/',
                url_text='Error')))

        self.assertSpyNotCalled(integration.start_build)

        # Check the status update created on publish.
        status_updates = StatusUpdate.objects.all()
        self.assertEqual(len(status_updates), 1)

        status_updates[0].run()

        self.assertSpyCalled(integration.start_build)
        call_kwargs = integration.start_build.last_call.kwargs
        status_update = call_kwargs['status_update']

        self.assertEqual(status_update, status_updates[0])
        self.assertAttrsEqual(
            status_update,
            {
                'description': 'some error.',
                'state': StatusUpdate.ERROR,
                'timeout': None,
                'url': 'https://example.com/error/',
                'url_text': 'Error',
            })

    @add_fixtures(['test_scmtools'])
    def test_manual_run_with_exception(self) -> None:
        """Testing BaseCIIntegration._on_status_update_request_run with
        exception
        """
        integration = self.integration

        self._setup_run_test(
            with_manual_run=True,
            spy_op=kgb.SpyOpRaise(Exception('oh no.')))

        self.assertSpyNotCalled(integration.start_build)

        # Check the status update created on publish.
        status_updates = StatusUpdate.objects.all()
        self.assertEqual(len(status_updates), 1)

        status_updates[0].run()

        self.assertSpyCalled(integration.start_build)
        call_kwargs = integration.start_build.last_call.kwargs
        status_update = call_kwargs['status_update']

        self.assertEqual(status_update, status_updates[0])
        self.assertAttrsEqual(
            status_update,
            {
                'description': 'internal error: oh no.',
                'state': StatusUpdate.ERROR,
                'timeout': None,
            })

    @add_fixtures(['test_scmtools'])
    def test_manual_run_without_matching_configs(self) -> None:
        """Testing BaseCIIntegration._on_status_update_request_run with
        matching configuration
        """
        integration = self.integration

        info = self._setup_run_test(
            with_manual_run=True,
            spy_op=kgb.SpyOpRaise(Exception('oh no.')))
        config = info['config']
        assert config

        self.assertSpyNotCalled(integration.start_build)

        # Check the status update created on publish.
        status_updates = StatusUpdate.objects.all()
        self.assertEqual(len(status_updates), 1)

        # Delete the old configuration.
        config.delete()
        get_integration_manager().clear_all_configs_cache()

        # Now attempt a manual run.
        status_updates[0].run()

        self.assertSpyNotCalled(integration.start_build)

    @add_fixtures(['test_scmtools'])
    def test_manual_run_without_diffset(self) -> None:
        """Testing BaseCIIntegration._on_status_update_request_run without
        DiffSet
        """
        integration = self.integration

        # We'll create with a DiffSet, and then delete it, to simulate
        # the conditions we check for in the code.
        info = self._setup_run_test(
            with_manual_run=True,
            spy_op=kgb.SpyOpRaise(Exception('oh no.')))
        info['diffsets'][0].delete()

        self.assertSpyNotCalled(integration.start_build)

        # Check the status update created on publish.
        status_updates = StatusUpdate.objects.all()
        self.assertEqual(len(status_updates), 1)

        # Now attempt a manual run.
        status_updates[0].run()

        self.assertSpyNotCalled(integration.start_build)

    @add_fixtures(['test_scmtools'])
    def test_manual_run_without_changedesc_diffset(self) -> None:
        """Testing BaseCIIntegration._on_status_update_request_run without
        new diff in review request update
        """
        integration = self.integration

        # We'll create with a DiffSet, and then delete the information in the
        # change description, to simulate the conditions we check for in the
        # code.
        #
        # This shouldn't be able to happen, but the code safe-guards against
        # this, so we're verifying those checks.
        info = self._setup_run_test(
            with_changedesc=True,
            with_manual_run=True,
            spy_op=kgb.SpyOpRaise(Exception('oh no.')))

        self.assertSpyNotCalled(integration.start_build)

        # Delete the 'diff' key.
        changedesc = info['review_request'].changedescs.all()[0]
        del changedesc.fields_changed['diff']
        changedesc.save(update_fields=('fields_changed',))

        # Check the status update created on publish.
        status_updates = StatusUpdate.objects.all()
        self.assertEqual(len(status_updates), 2)

        # Now attempt a manual run.
        status_updates[1].run()

        self.assertSpyNotCalled(integration.start_build)

    @add_fixtures(['test_scmtools'])
    def test_manual_run_without_service_id_match(self) -> None:
        """Testing BaseCIIntegration._on_status_update_request_run without
        service ID match
        """
        integration = self.integration

        self._setup_run_test(with_manual_run=True)

        self.assertSpyNotCalled(integration.start_build)

        # Check the status update created on publish.
        status_updates = StatusUpdate.objects.all()
        self.assertEqual(len(status_updates), 1)

        status_update = status_updates[0]

        # Set a new service ID and run it.
        status_update.service_id = 'other-service'
        status_update.run()

        self.assertSpyNotCalled(integration.start_build)

    @add_fixtures(['test_scmtools'])
    def test_timeout(self) -> None:
        """Testing BaseCIIntegration._on_status_update_request_run with
        timeout
        """
        integration = self.integration

        info = self._setup_run_test(with_manual_run=True,
                                    timeout=12345)
        config = info['config']
        diffsets = info['diffsets']
        review_request = info['review_request']

        assert config

        self.assertSpyNotCalled(integration.start_build)

        # Check the status update created on publish.
        status_update = StatusUpdate.objects.get()
        user = integration.get_or_create_user()

        self.assertIsNotNone(status_update.pk)
        self.assertAttrsEqual(
            status_update,
            {
                'change_description': None,
                'description': 'waiting to run.',
                'extra_data': {
                    '__integration_config_id': config.pk,
                    'can_retry': True,
                },
                'review_request': review_request,
                'service_id': 'my-ci',
                'state': StatusUpdate.NOT_YET_RUN,
                'summary': 'My CI',
                'timeout': 12345,
                'user': user,
            })

        # Perform the manual run.
        status_update.run()

        self.assertSpyCalled(integration.start_build)
        call_kwargs = integration.start_build.last_call.kwargs
        prep_data = call_kwargs['prep_data']

        self.assertEqual(call_kwargs['status_update'], status_update)
        self.assertEqual(call_kwargs['config'], config)

        status_update = call_kwargs['status_update']

        self.assertIsNotNone(status_update.pk)
        self.assertAttrsEqual(
            status_update,
            {
                'change_description': None,
                'description': 'starting build...',
                'extra_data': {
                    '__integration_config_id': config.pk,
                    'can_retry': True,
                },
                'review_request': review_request,
                'service_id': 'my-ci',
                'state': StatusUpdate.PENDING,
                'summary': 'My CI',
                'timeout': 12345,
                'user': user,
            })

        self.assertAttrsEqual(
            prep_data,
            {
                'changedesc': None,
                'configs': [config],
                'diffset': diffsets[0],
                'extra_state': {
                    'my_state': 123,
                },
                'review_request': review_request,
            })
        self.assertEqual(prep_data.user.username, 'test-ci-user')

        self.assertIsNotNone(status_update.pk)
        self.assertAttrsEqual(
            status_update,
            {
                'change_description': None,
                'description': 'starting build...',
                'extra_data': {
                    '__integration_config_id': config.pk,
                    'can_retry': True,
                },
                'review_request': review_request,
                'service_id': 'my-ci',
                'state': StatusUpdate.PENDING,
                'summary': 'My CI',
                'timeout': 12345,
                'user': prep_data.user,
            })

    def _setup_run_test(
        self,
        *,
        with_changedesc: bool = False,
        with_config: bool = True,
        with_diffset: bool = True,
        with_manual_run: bool = False,
        timeout: Optional[int] = None,
        spy_op: Optional[BaseSpyOperation] = None,
    ) -> SetupRunTestResults:
        """Set up a test for a build run, and handle publishing.

        This will create a configuration, review request, diffset, and any
        other necessary objects for a test for a build run. It also takes care
        of the spying and the publishing.

        Args:
            with_changedesc (bool, optional):
                Whether to test a publish on a change description.

            with_config (bool, optional):
                Whether to create a configuration to test with.

            with_diffset (bool, optional):
                Whether to create diffsets on the review request or change
                description.

            with_manual_run (bool, optional):
                Whether to test with a manual run configuration.

            timeout (int, optional):
                Optional number of seconds until the build is considered to
                have timed out.

                Version Added:
                    4.0

            spy_op (kgb.ops.BaseSpyOperation, optional):
                An explicit spy operation to use for checking builds.

        Returns:
            dict:
            A dictionary of results. See :py:class:`SetupRunTestResults`.
        """
        diffsets: List[DiffSet] = []

        integration = self.integration
        integration.enable_integration()

        if not with_changedesc:
            # We're testing the initial publish, so spy on it here.
            self.spy_on(integration.start_build, op=spy_op)

        if with_config:
            config = self._create_config(name='config',
                                         with_manual_run=with_manual_run,
                                         timeout=timeout)
        else:
            config = None

        review_request = self.create_review_request(
            create_repository=True,
            target_people=[self.create_user()])

        if with_diffset:
            diffsets.append(self.create_diffset(
                review_request,
                revision=1,
                draft=True))

        review_request.publish(user=review_request.owner)

        if with_changedesc:
            # We haven't spied above. We want to spy now, before the next
            # publish.
            self.spy_on(integration.start_build, op=spy_op)

            draft = self.create_review_request_draft(review_request)
            draft.summary = 'New summary'
            draft.save(update_fields=('summary',))

            if with_diffset:
                diffsets.append(self.create_diffset(
                    review_request,
                    revision=2,
                    draft=True))

            review_request.publish(user=review_request.owner)

        return {
            'config': config,
            'diffsets': diffsets,
            'review_request': review_request,
        }

    def _create_config(
        self,
        *,
        name: str,
        integration_id: str = MyCIIntegration.integration_id,
        enabled: bool = True,
        timeout: Optional[int] = None,
        local_site: Optional[LocalSite] = None,
        with_condition_match: bool = True,
        with_manual_run: bool = False,
    ) -> IntegrationConfig:
        """Create an integration config for testing.

        Args:
            name (str):
                The name of the configuration.

            integration_id (str, optional):
                The ID of the integration the configuration applies to.

            enabled (bool, optional):
                Whether this configuration is enabled.

            timeout (int, optional):
                Optional number of seconds until the build is considered to
                have timed out.

                Version Added:
                    4.0

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site the configuration is bound to.

            with_condition_match (bool, optional):
                Whether the condition should match the test run.

            with_manual_run (bool, optional):
                Whether the configuration should be created in manual run mode.

        Returns:
            reviewboard.integrations.models.IntegrationConfig:
            The resulting integration configuration.
        """
        choice = ReviewRequestRepositoriesChoice()

        if with_condition_match:
            operator = choice.get_operator('any')
        else:
            operator = choice.get_operator('none')

        condition_set = ConditionSet(conditions=[
            Condition(choice=choice,
                      operator=operator)
        ])

        config_settings: Dict[str, Any] = {
            'conditions': condition_set.serialize(),
        }

        if with_manual_run:
            config_settings['run_manually'] = True

        if timeout is not None:
            config_settings['timeout'] = timeout

        return IntegrationConfig.objects.create(
            integration_id=integration_id,
            local_site=local_site,
            name=name,
            enabled=enabled,
            settings=config_settings)
